from ontobio.sim.api.interfaces import SimApi
from ontobio.config import get_config
from typing import Optional, List
from ontobio.vocabulary.similarity import SimAlgorithm
from ontobio.model.similarity import SimResult


class OwlSim3Api(SimApi):
    """
    OwlSim3 is the next iteration of owlsim and contains more algorithms
    than owlsim2 including several variations of the BOQA algorithm

    Ref:
    https://github.com/monarch-initiative/owlsim-v3
    http://owlsim3.monarchinitiative.org/api/docs/

    Status: not implemented
    """

    def __init__(self, url: Optional[str]=None, timeout: Optional[int]=None):
        self.url = url if url is not None else get_config().owlsim3.url
        self.timeout = timeout if timeout is not None else get_config().owlsim3.timeout

    def search(self,
               id_list: List,
               negated_classes: List,
               limit: Optional[int],
               method: Optional) -> List[SimResult]:
        """
        Given an input list of classes or individuals,
        provides a ranking of similar profiles
        """
        raise NotImplementedError

    def compare(self,
                query_classes: List,
                reference_classes: List,
                method: Optional) -> SimResult:
        """
        Given two lists of entites (classes, individual)
        return their similarity
        """
        raise NotImplementedError

    @staticmethod
    def matchers() -> List[SimAlgorithm]:
        """
        Matchers in owlsim3
        Could be dynamically retrieved from
        http://owlsim3.monarchinitiative.org/api/match/matchers
        """
        return [
            SimAlgorithm.PHENODIGM,
            SimAlgorithm.BAYES_NETWORK,
            SimAlgorithm.BAYES_VARIABLE,
            SimAlgorithm.NAIVE_BAYES_THREE_STATE,
            SimAlgorithm.NAIVE_BAYES_TWO_STATE,
            SimAlgorithm.NAIVE_BAYES_TWO_STATE_NO_BLANKET,
            SimAlgorithm.GRID,
            SimAlgorithm.GRID_NEGATED,
            SimAlgorithm.JACCARD,
            SimAlgorithm.MAX_INFORMATION
        ]

    def __str__(self):
        return "owlsim3 api: {}".format(self.url)



"""
Mapping between native ontology model and rdflib
"""
from ontobio.ontol import Ontology, Synonym
from prefixcommons.curie_util import contract_uri, expand_uri, get_prefixes
from rdflib import Namespace
from rdflib import BNode
from rdflib import Literal
from rdflib import URIRef
from rdflib.namespace import RDFS
from rdflib.namespace import OWL
import networkx
import logging

logger = logging.getLogger(__name__)


def rdfgraph_to_ontol(rg):
    """
    Return an Ontology object from an rdflib graph object

    Status: Incomplete
    """
    digraph = networkx.MultiDiGraph()
    from rdflib.namespace import RDF
    label_map = {}
    for c in rg.subjects(RDF.type, OWL.Class):
        cid = contract_uri_wrap(c)
        logger.info("C={}".format(cid))
        for lit in rg.objects(c, RDFS.label):
            label_map[cid] = lit.value
            digraph.add_node(cid, label=lit.value)
        for s in rg.objects(c, RDFS.subClassOf):
            # todo - blank nodes
            sid = contract_uri_wrap(s)
            digraph.add_edge(sid, cid, pred='subClassOf')

    logger.info("G={}".format(digraph))
    payload = {
        'graph': digraph,
        #'xref_graph': xref_graph,
        #'graphdoc': obographdoc,
        #'logical_definitions': logical_definitions
    }
            
    ont = Ontology(handle='wd', payload=payload)
    return ont



                
        
def contract_uri_wrap(uri):
    curies = contract_uri(uri)
    if len(curies) > 0:
        return curies[0]
    else:
        return uri




__relation_label_lookup = {
    "occurs in": "http://purl.obolibrary.org/obo/BFO_0000066",
    "happens during": "http://purl.obolibrary.org/obo/RO_0002092",
    "has input": "http://purl.obolibrary.org/obo/RO_0002233",
    "results in specification of": "http://purl.obolibrary.org/obo/RO_0002356",
    "part of": "http://purl.obolibrary.org/obo/BFO_0000050",
    "has part": "http://purl.obolibrary.org/obo/BFO_0000051",
    "results in development of": "http://purl.obolibrary.org/obo/RO_0002296",
    "results in movement of": "http://purl.obolibrary.org/obo/RO_0002565",
    "occurs at": "http://purl.obolibrary.org/obo/GOREL_0000501",
    "stabilizes": "http://purl.obolibrary.org/obo/GOREL_0000018",
    "positively regulates": "http://purl.obolibrary.org/obo/RO_0002213",
    "regulates transport of": "http://purl.obolibrary.org/obo/RO_0002011",
    "regulates transcription of": "http://purl.obolibrary.org/obo/GOREL_0098788",
    "causally upstream of": "http://purl.obolibrary.org/obo/RO_0002411",
    "regulates activity of": "http://purl.obolibrary.org/obo/GOREL_0098702",
    "adjacent to": "http://purl.obolibrary.org/obo/RO_0002220",
    "results in acquisition of features of": "http://purl.obolibrary.org/obo/RO_0002315",
    "results in morphogenesis of": "http://purl.obolibrary.org/obo/RO_0002298",
    "results in maturation of": "http://purl.obolibrary.org/obo/RO_0002299",
    "has participant": "http://purl.obolibrary.org/obo/RO_0000057",
    "transports or maintains localization of": "http://purl.obolibrary.org/obo/RO_0002313",
    "negatively regulates": "http://purl.obolibrary.org/obo/RO_0002212",
    "regulates": "http://purl.obolibrary.org/obo/RO_0002211",
    "regulates expression of": "http://purl.obolibrary.org/obo/GOREL_0098789",
    "has target end location": "http://purl.obolibrary.org/obo/RO_0002339",
    "produced by": "http://purl.obolibrary.org/obo/RO_0003001",
    "has end location": "http://purl.obolibrary.org/obo/RO_0002232",
    "directly positively regulates": "http://purl.obolibrary.org/obo/RO_0002629",
    "has direct input": "http://purl.obolibrary.org/obo/GOREL_0000752",
    "enables": "http://purl.obolibrary.org/obo/RO_0002327",
    "enabled by": "http://purl.obolibrary.org/obo/RO_0002333",
    "involved in": "http://purl.obolibrary.org/obo/RO_0002331",
    "acts upstream of": "http://purl.obolibrary.org/obo/RO_0002263",
    "colocalizes with": "http://purl.obolibrary.org/obo/RO_0002325",
    "contributes to": "http://purl.obolibrary.org/obo/RO_0002326"
}

def lookup_label(label, default=None):
    return __relation_label_lookup.get(label.replace("_", " "), default)

def label_relation_lookup():
    return __relation_label_lookup



import json

from ontobio.io import gaference
from ontobio.model import association


def test_build_annotation_inferences():
    with open("tests/resources/test.inferences.json") as inferences_file:
        gaferences = json.load(inferences_file)
        inferences = gaference.build_annotation_inferences(gaferences)

        akey = gaference.AnnotationKey(
            gaference.RelationTo("http://purl.obolibrary.org/obo/BFO_0000050", "http://purl.obolibrary.org/obo/GO_0036064"),
            "http://purl.obolibrary.org/obo/NCBITaxon_10090",
            association.ExtensionConjunctions(frozenset([
                association.ExtensionUnit("http://purl.obolibrary.org/obo/BFO_0000050", "http://purl.obolibrary.org/obo/EMAPA_17168"),
                association.ExtensionUnit("http://purl.obolibrary.org/obo/BFO_0000050", "http://purl.obolibrary.org/obo/CL_0010009")
            ])))

        val = inferences[akey]

        expected = gaference.InferenceValue(True, False,
            [gaference.RelationTo("http://purl.obolibrary.org/obo/BFO_0000050", "http://purl.obolibrary.org/obo/GO_0097458")])

        assert val == expected


def test_produce_inference_produces_inferences():
    with open("tests/resources/test.inferences.json") as inference_file:
        gaferences = json.load(inference_file)
        inferences = gaference.build_annotation_inferences(gaferences)

    gaf_line = "MGI\tMGI:2178217\tAkap9\t\tGO:0036064\tMGI:MGI:5303017|PMID:22031837\tIDA\t\tC\tA kinase (PRKA) anchor protein (yotiao) 9\t5730481H23Rik|AKAP450|G1-448-15|mei2-5|repro12\tprotein\ttaxon:10090\t20131226\tMGI\tpart_of(EMAPA:17168),part_of(CL:0010009)\t"
    gaf = gaf_line.split("\t")

    results = gaference.produce_inferences(gaf, inferences)
    expected_line = ["MGI", "MGI:2178217", "Akap9", "", "GO:0097458", "MGI:MGI:5303017|PMID:22031837", "IDA", "", "C", "A kinase (PRKA) anchor protein (yotiao) 9", "5730481H23Rik|AKAP450|G1-448-15|mei2-5|repro12", "protein", "taxon:10090", "20131226", "MGI", "", ""]
    assert len(results) == 1
    assert len(results[0].inferred_gafs) == 1
    assert results[0].inferred_gafs[0] == expected_line

def test_produce_inference_produces_many_inferences():
    with open("tests/resources/test.inferences.json") as inference_file:
        gaferences = json.load(inference_file)
        inferences = gaference.build_annotation_inferences(gaferences)

    gaf_line = "MGI\tMGI:1345162\tAdam23\t\tGO:0099056\tMGI:MGI:4431144|PMID:20133599\tIDA\t\tC\ta disintegrin and metallopeptidase domain 23\tMDC3\tprotein\ttaxon:10090\t20180711\tSynGO\tpart_of(GO:0098978),part_of(UBERON:0000061),part_of(EMAPA:35405)|part_of(GO:0098978),part_of(UBERON:0000061),part_of(EMAPA:16894)\t"
    gaf = gaf_line.split("\t")

    results = gaference.produce_inferences(gaf, inferences)
    expected_lines = [
        ["MGI", "MGI:1345162", "Adam23", "", "GO:0098978", "MGI:MGI:4431144|PMID:20133599", "IDA", "", "C", "a disintegrin and metallopeptidase domain 23", "MDC3", "protein", "taxon:10090", "20180711", "SynGO", "", ""],
        ["MGI", "MGI:1345162", "Adam23", "", "GO:0098978", "MGI:MGI:4431144|PMID:20133599", "IDA", "", "C", "a disintegrin and metallopeptidase domain 23", "MDC3", "protein", "taxon:10090", "20180711", "SynGO", "", ""],
    ]
    assert len(results) == 2
    assert results[0].inferred_gafs[0] == expected_lines[0]
    assert results[1].inferred_gafs[0] == expected_lines[1]


def test_taxon_check_failure():
    with open("tests/resources/test.inferences.json") as inference_file:
        gaferences = json.load(inference_file)
        inferences = gaference.build_annotation_inferences(gaferences)

    gaf_line = "MGI\tMGI:1924956\tAbcb5\t\tGO:0048058\tMGI:MGI:5585659|PMID:25030174\tIMP\t\tP\tATP-binding cassette, sub-family B (MDR/TAP), member 5\t9230106F14Rik\tprotein\ttaxon:10090\t20140729\tUniProt\t\t"
    gaf = gaf_line.split("\t")

    results = gaference.produce_inferences(gaf, inferences)

    assert len(results) == 1
    assert results[0].problem == gaference.ProblemType.TAXON

def test_pombase_taxon_failure():
    with open("tests/resources/test.inferences.json") as inference_file:
        gaferences = json.load(inference_file)
        inferences = gaference.build_annotation_inferences(gaferences)

    gaf_line = "PomBase\tSPBC11B10.09\tcdc2\t\tGO:0007275\tPMID:21873635\tIBA\tPANTHER:PTN000623979|TAIR:locus:2099478\tP\tCyclin-dependent kinase 1\tUniProtKB:P04551|PTN000624043\tprotein\ttaxon:284812\t20170228\tGO_Central"
    gaf = gaf_line.split("\t")

    results = gaference.produce_inferences(gaf, inferences)

    assert len(results) == 1
    assert results[0].problem == gaference.ProblemType.TAXON

def test_extension_check_failure():
    with open("tests/resources/test.inferences.json") as inference_file:
        gaferences = json.load(inference_file)
        inferences = gaference.build_annotation_inferences(gaferences)

    gaf_line = "MGI\tMGI:109192\tActn2\t\tGO:0072659\tMGI:MGI:4366185|PMID:19815520\tIMP\t\tP\tactinin alpha 2\t1110008F24Rik\tprotein\ttaxon:10090\t20150506\tUniProt\tpart_of(CL:0002495),has_direct_input(UniProtKB:P58390)\t"
    gaf = gaf_line.split("\t")

    results = gaference.produce_inferences(gaf, inferences)

    assert len(results) == 1
    assert results[0].problem == gaference.ProblemType.EXTENSION



from ontobio.io.entityparser import BgiParser
from ontobio.assoc_factory import AssociationSetFactory
from ontobio.ontol_factory import OntologyFactory

BGI = "tests/resources/fb-bgi.json"
    
def test_parse_gpi():
    p = BgiParser()
    results = p.parse(open(BGI,"r"))
    for r in results:
        print(r)
    r1 = results[0]
    assert r1['label'] == '4.5SRNA'
    assert r1['taxon']['id'] == 'NCBITaxon:7227'
    assert r1['full_name'] == '4.5SRNA'

    [r] = [r for r in results if r['full_name'] == 'Acetylcholine esterase']
    
    assert 'UniProtKB:P07140' in r['xrefs']
    assert "ace-2" in r['synonyms']

    for m in p.report.messages:
        print("MESSAGE: {}".format(m))
    assert len(p.report.messages) == 0
    print(p.report.to_markdown())
    
    



"""
An ontology backed by a remote Wikidata SPARQL service.
"""

import networkx as nx
import logging
import ontobio.ontol
from ontobio.ontol import Ontology, Synonym
from prefixcommons.curie_util import contract_uri, expand_uri, get_prefixes
from SPARQLWrapper import SPARQLWrapper, JSON, RDF
from ontobio.sparql.rdflib_bridge import rdfgraph_to_ontol
from ontobio.sparql.sparql_ontol_utils import OIO_SYNS, get_named_graph, run_sparql

logger = logging.getLogger(__name__)


class WikidataOntology(Ontology):
    """
    An ontology backed by a remote Wikidata SPARQL service.
    """


    # TODO
    # Override
    def resolve_names(self, names, is_remote=False, synonyms=False, **args):
        if not is_remote:
            # TODO: ensure synonyms present
            return super().resolve_names(names, **args)
        else:
            results = set()
            for name in names:
                results.update( self._search(name, 'rdfs:label', **args) )
                if synonyms:
                    for pred in OIO_SYNS.values():
                        results.update( self._search(name, pred, **args) )
            logger.info("REMOTE RESULTS="+str(results))
            return list(results)

    # TODO
    def _search(self, searchterm, pred, **args):
        """
        Search for things using labels
        """
        # TODO: DRY with sparql_ontol_utils
        searchterm = searchterm.replace('%','.*')
        namedGraph = get_named_graph(self.handle)
        query = """
        prefix oboInOwl: <http://www.geneontology.org/formats/oboInOwl#>
        SELECT ?c WHERE {{
        GRAPH <{g}>  {{
        ?c {pred} ?l
        FILTER regex(?l,'{s}','i')
        }}
        }}
        """.format(pred=pred, s=searchterm, g=namedGraph)
        bindings = run_sparql(query)
        return [r['c']['value'] for r in bindings]

    # TODO
    def sparql(self, select='*', body=None, inject_prefixes=None, single_column=False):
        """
        Execute a SPARQL query.

        The query is specified using `select` and `body` parameters.
        The argument for the Named Graph is injected into the query.

        The select parameter should be either '*' or a list of vars (not prefixed with '?').

         - If '*' is passed, then the result is a list of dicts, { $var: {value: $val } }
         - If a list of vars is passed, then the result is a list of lists
         - Unless single_column=True, in which case the results are a simple list of values from the first var

        The inject_prefixes argument can be used to inject a list of prefixes - these are expanded
        using the prefixcommons library
        
        """
        if inject_prefixes is None:
            inject_prefixes = []
        namedGraph = get_named_graph(self.handle)
        cols = []
        select_val = None
        if select is None or select=='*':
            if not single_column:
                cols=None
            select_val='*'
        else:
            if isinstance(cols,list):
                cols = [select]
            else:
                cols = select
            select_val = ", ".join(['?'+c for c in cols])

        prefixes = ""
        if inject_prefixes is not None:
            plist = ["prefix {}: <{}> ".format(p,expand_uri(p+":")) for p in inject_prefixes if p != "" and p is not None]
            prefixes = "\n".join(plist)
        query = """
        {prefixes}
        SELECT {s} WHERE {{
        GRAPH <{g}>  {{
        {b}
        }}
        }}
        """.format(prefixes=prefixes, s=select_val, b=body, g=namedGraph)
        bindings = run_sparql(query)
        if len(bindings) == 0:
            return []
        if cols is None:
            return bindings
        else:
            if single_column:
                c = list(bindings[0].keys())[0]
                return [r[c]['value'] for r in bindings]
            else:
                return [r[c]['value'] for c in cols for r in bindings]
    
    
class EagerWikidataOntology(WikidataOntology):
    """
    Eager Wikidata ontology implementation.

    Caches all classes that are sub or super classes of a specified node
    """

    def __init__(self, handle=None):
        """
        initializes based on an ontology class ID (Q ID) in wikidata
        """
        handle = handle.replace('wdq:','')
        self.handle = handle
        logger.info("Creating eager-wikidata-ont from "+str(handle))
        self.create_from_hub(handle)

    def __str__(self):
        return "h:{} g:{}".format(self.handle, self.graph)

    def create_from_hub(self, hub_id):
        if hub_id.find(":") == -1:
            hub_id = 'wd:' + hub_id
        if hub_id.startswith('http'):
            hub_id = '<' + hub_id + '>'
        q = """
        PREFIX wd: <http://www.wikidata.org/entity/> 
        CONSTRUCT {{
        ?cls rdfs:label ?clsLabel .
        ?cls a owl:Class .
        ?superClass a owl:Class .
        ?superClass rdfs:label ?superClassLabel .
        ?cls rdfs:subClassOf ?superClass
        }}
        WHERE {{
         {{  {{ ?cls wdt:P279* {hub_id} }} UNION
             {{ {hub_id} wdt:P279* ?cls }} }}
        ?cls wdt:P279 ?superClass
        SERVICE wikibase:label {{
          bd:serviceParam wikibase:language "en" .
        }}
        }}
        """.format(hub_id=hub_id)
        logger.info("QUER={}".format(q))
        sparql = SPARQLWrapper("http://query.wikidata.org/sparql")
        sparql.setQuery(q)
        sparql.setReturnFormat(RDF)
        rg = sparql.query().convert()
        logger.info("RG={}".format(rg))
        ont = rdfgraph_to_ontol(rg)
        self.graph = ont.graph
        

class LazyWikidataOntology(WikidataOntology):
    """
    Non-caching wikidata-backed ontology TODO
    """

    def __init__(self):
        self.all_logical_definitions = [] ## TODO


    



"""
Classes for exporting associations.

"""
import re
import datetime
import json

from typing import List

from ontobio import ecomap

external_taxon = re.compile("taxon:([0-9]+)")
internal_taxon = re.compile("NCBITaxon:([0-9]+)")

def _str(v):
    if v is None:
        return ""
    else:
        return str(v)

class AssocWriterConfig():
    """
    Placeholder class for configuration object for all writers
    """
    pass

class AssocWriter():
    """
    Abstract superclass of all association writer objects (Gpad, GAF)
    """
    def _split_prefix(self, ref):
        id = ref['id']
        [prefix, local_id] = id.split(':', maxsplit=1)
        return prefix, local_id

    def _write_row(self, vals):
        line = self.tsv_as_string(vals)
        if self.file:
            self.file.write(line+"\n")
        else:
            print(line)
            
    def tsv_as_string(self, vals) -> str:
        return "\t".join([_str(v) for v in vals])

    def _write(self, line):
        if self.file:
            self.file.write(line)
        else:
            print(line)

    def _extension_expression(self, object_extensions):
        unions = []
        for union_key, union_value in object_extensions.items():
            if union_key == "union_of":
                # union_value is list of { "intersection_of" ...}
                for union_item in union_value:
                    for intersect_key, intersect_value in union_item.items():
                        if intersect_key == "intersection_of":
                            intersections = []
                            for intersect_item in intersect_value:
                                prop = intersect_item["property"]
                                filler = intersect_item["filler"]
                                full_property = "{property}({filler})".format(property=prop, filler=filler)
                                intersections.append(full_property)
                            joined_intersections = ",".join(intersections)
                            unions.append(joined_intersections)

        return "|".join(unions)

    def normalize_taxon(self, taxon):
        global internal_taxon
        global external_taxon

        if taxon == None:
            return ""

        if external_taxon.match(taxon):
            # If we match here, then the internal view already exists and we're good
            return internal_taxon

        match = internal_taxon.match(taxon)
        if match:
            taxon_id = match.group(1)
            return "taxon:{num}".format(num=taxon_id)

        return taxon
        
    def as_tsv(self, assoc) -> List[str]:
        """
        Transform a single association to a string line.
        """
        pass

    def write_assoc(self, assoc):
        """
        Write a single association to a line in the output file
        """
        vals = self.as_tsv(assoc)
        if vals != []:
            # Write it if we found content
            self._write_row(vals)

    def write(self, assocs, meta=None):
        """
        Write a complete set of associations to a file

        Arguments
        ---------
        assocs: list[dict]
            A list of association dict objects
        meta: Meta
            metadata about association set (not yet implemented)

        """
        for a in assocs:
            self.write_assoc(a)

class GpadWriter(AssocWriter):
    """
    Writes Associations in GPAD format
    """
    def __init__(self, file=None):
        self.file = file
        self._write("!gpa-version: 1.1\n")
        self.ecomap = ecomap.EcoMap()

    def as_tsv(self, assoc):
        """
        Write a single association to a line in the output file
        """
        if assoc.get("header", False):
            return []

        subj = assoc['subject']

        db, db_object_id = self._split_prefix(subj)

        rel = assoc['relation']
        qualifier = rel['id']
        if assoc['negated']:
            qualifier = 'NOT|' + qualifier

        goid = assoc['object']['id']

        ev = assoc['evidence']

        evidence = self.ecomap.coderef_to_ecoclass(ev['type'])
        withfrom = "|".join(ev['with_support_from'])
        reference = "|".join(ev['has_supporting_reference'])

        date = assoc['date']
        assigned_by = assoc['provided_by']

        annotation_properties = '' # TODO
        interacting_taxon_id = assoc['interacting_taxon']
        vals = [db,
                db_object_id,
                qualifier,
                goid,
                reference,
                evidence,
                withfrom,
                interacting_taxon_id, # TODO
                date,
                assigned_by,
                self._extension_expression(assoc['object_extensions']),
                annotation_properties]

        return vals

class GafWriter(AssocWriter):
    """
    Writes Associations in GAF format. Not yet implemented
    """
    def __init__(self, file=None, source=None):
        self.file = file
        self._write("!gaf-version: 2.1\n")
        self._write("!\n")
        self._write("!Generated by GO Central\n")
        self._write("!\n")
        self._write("!Date Generated by GOC: {}\n".format(str(datetime.date.today())))
        self._write("!\n")
        # Just uses the word `source` if source is none. Otherwise uses the name of the source in the header
        self._write("!Header from {source}source association file:\n".format(source=source+" " if source else ""))
        self._write("!=================================\n")

    def _full_taxon_field(self, taxon, interacting_taxon):
        full_taxon = taxon
        if interacting_taxon not in [None, ""]:
            full_taxon = "{taxon}|{interacting_taxon}".format(taxon=taxon, interacting_taxon=interacting_taxon)

        return full_taxon

    def as_tsv(self, assoc):
        """
        Write a single association to a line in the output file
        """
        # Handle comment 'associations'
        if assoc.get("header", False):

            # Skip incoming gaf-version headers, as we created the version above already
            if re.match("![\s]*gaf.?version", assoc["line"]):
                return []

            self._write(assoc["line"] + "\n")
            return []

        # print("Writing assoc {}".format(assoc))
        subj = assoc['subject']

        db, db_object_id = self._split_prefix(subj)

        qualifier = "|".join(assoc["qualifiers"])
        # qualifier is | seperated string, filter on None removes empty strings in case qualifiers is empty
        if assoc['negated']:
            qualifier = "|".join(list(filter(None, ["NOT", qualifier])))

        goid = assoc['object']['id']

        ev = assoc['evidence']
        evidence = ev['type']
        withfrom = "|".join(ev['with_support_from'])
        reference = "|".join(ev['has_supporting_reference'])

        date = assoc['date']
        assigned_by = assoc['provided_by']

        annotation_properties = '' # TODO
        # if we have any subject extensions, list each one that has a "property" equal to "isoform", take the first one, and grab the "filler"
        gene_product_isoform = [e for e in assoc["subject_extensions"] if e["property"] == "isoform"][0]["filler"] if len(assoc["subject_extensions"]) > 0 else ""

        aspect = assoc['aspect']
        interacting_taxon_id = assoc["interacting_taxon"]
        taxon = self._full_taxon_field(self.normalize_taxon(subj['taxon']['id']), self.normalize_taxon(interacting_taxon_id))

        extension_expression = self._extension_expression(assoc['object_extensions'])

        vals = [db,
                db_object_id,
                subj.get('label'),
                qualifier,
                goid,
                reference,
                evidence,
                withfrom,
                aspect,
                subj["fullname"],
                "|".join(subj.get('synonyms',[])),
                subj.get('type'),
                taxon,
                date,
                assigned_by,
                extension_expression,
                gene_product_isoform]

        return vals



from ontobio.ontol_factory import OntologyFactory


class GoAspector:
    def __init__(self, go_ontology):
        if go_ontology:
            self.ontology = go_ontology
        else:
            self.ontology = OntologyFactory().create("go")

    def get_ancestors_through_subont(self, go_term, relations):
        """
        Returns the ancestors from the relation filtered GO subontology of go_term's ancestors.

        subontology() primarily used here for speed when specifying relations to traverse. Point of this is to first get
        a smaller graph (all ancestors of go_term regardless of relation) and then filter relations on that instead of
        the whole GO.
        """
        all_ancestors = self.ontology.ancestors(go_term, reflexive=True)
        subont = self.ontology.subontology(all_ancestors)
        return subont.ancestors(go_term, relations)

    def get_isa_partof_closure(self, go_term):
        return self.get_ancestors_through_subont(go_term, relations=["subClassOf", "BFO:0000050"])

    def get_isa_closure(self, go_term):
        return self.get_ancestors_through_subont(go_term, relations=["subClassOf"])

    def is_biological_process(self, go_term):
        """
        Returns True is go_term has is_a, part_of ancestor of biological process GO:0008150
        """
        bp_root = "GO:0008150"
        if go_term == bp_root:
            return True
        ancestors = self.get_isa_closure(go_term)
        if bp_root in ancestors:
            return True
        else:
            return False

    def is_molecular_function(self, go_term):
        """
        Returns True is go_term has is_a, part_of ancestor of molecular function GO:0003674
        """
        mf_root = "GO:0003674"
        if go_term == mf_root:
            return True
        ancestors = self.get_isa_closure(go_term)
        if mf_root in ancestors:
            return True
        else:
            return False

    def is_cellular_component(self, go_term):
        """
        Returns True is go_term has is_a, part_of ancestor of cellular component GO:0005575
        """
        cc_root = "GO:0005575"
        if go_term == cc_root:
            return True
        ancestors = self.get_isa_closure(go_term)
        if cc_root in ancestors:
            return True
        else:
            return False

    def go_aspect(self, go_term):
        """
        For GO terms, returns F, C, or P corresponding to its aspect
        """
        if not go_term.startswith("GO:"):
            return None
        else:
            # Check ancestors for root terms
            if self.is_molecular_function(go_term):
                return 'F'
            elif self.is_cellular_component(go_term):
                return 'C'
            elif self.is_biological_process(go_term):
                return 'P'



from ontobio import __version__ as ontobio_version

def get_user_agent(name="ontobio", version=ontobio_version, modules=None, caller_name=None):
    """
    Create a User-Agent string
    """

    user_agent_array = ["{}/{}".format(name, version)]
    if modules:
        module_info_array = []
        for m in modules:
            mod_name = m.__name__
            mod_version = None
            if hasattr(m, 'get_version'):
                mod_version = m.get_version()
            else:
                mod_version = m.__version__
            module_info_array.append("{}/{}".format(mod_name, mod_version))

        if caller_name:
            module_info_array.append(caller_name)

        user_agent_array.append("({})".format('; '.join(module_info_array)))
    else:
        if caller_name:
            user_agent_array.append("({})".format(caller_name))

    return ' '.join(user_agent_array)



#!/usr/bin/env python

"""
Command line wrapper to obographs library.

Example:
(venv) ~/repos/biolink-api(master) $ python obographs/bin/phenolog.py -vvv -r obo:mp -R go 'abnormal cardiovascular system physiology'

With background:
python obographs/bin/phenolog.py -b 'nervous system phenotype' -v -r cache/ontologies/monarch.json -R go 'abnormal nervous system morphology'



"""

import argparse
import networkx as nx
from networkx.algorithms.dag import ancestors, descendants
from ontobio.ontol_factory import OntologyFactory
from ontobio.assoc_factory import AssociationSetFactory
from ontobio.graph_io import GraphRenderer
from ontobio.slimmer import get_minimal_subgraph
import logging
import sys


def main():
    """
    Phenologs
    """
    parser = argparse.ArgumentParser(description='Phenologs'
                                                 """
                                                 By default, ontologies are cached locally and synced from a remote sparql endpoint
                                                 """,
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-r', '--resource1', type=str, required=False,
                        help='Name of ontology1')
    parser.add_argument('-R', '--resource2', type=str, required=False,
                        help='Name of ontology2')
    parser.add_argument('-T', '--taxon', type=str, default='NCBITaxon:10090', required=False,
                        help='NCBITaxon ID')
    parser.add_argument('-s', '--search', type=str, default='', required=False,
                        help='Search type. p=partial, r=regex')
    parser.add_argument('-b', '--background', type=str, default=None, required=False,
                        help='Class to use for background')
    parser.add_argument('-p', '--pthreshold', type=float, default=0.05, required=False,
                        help='P-value threshold')
    parser.add_argument('-v', '--verbosity', default=0, action='count',
                        help='Increase output verbosity')

    parser.add_argument('ids',nargs='*')

    args = parser.parse_args()

    if args.verbosity >= 2:
        logging.basicConfig(level=logging.DEBUG)
    if args.verbosity == 1:
        logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)

    logger.info("Welcome!")

    ofactory = OntologyFactory()
    afactory = AssociationSetFactory()
    handle = args.resource1
    ont1 = ofactory.create(args.resource1)
    ont2 = ofactory.create(args.resource2)
    logger.info("onts: {} {}".format(ont1, ont2))
    searchp = args.search

    category = 'gene'

    aset1 = afactory.create(ontology=ont1,
                            subject_category=category,
                            object_category='phenotype',
                            taxon=args.taxon)
    aset2 = afactory.create(ontology=ont2,
                            subject_category=category,
                            object_category='function',
                            taxon=args.taxon)

    bg_cls = None
    if args.background is not None:
        bg_ids = resolve(ont1,[args.background],searchp)
        if len(bg_ids) == 0:
            logger.error("Cannnot resolve: '{}' using {} in {}".format(args.background, searchp, ont1))
            sys.exit(1)
        elif len(bg_ids) > 1:
            logger.error("Multiple matches: '{}' using {} MATCHES={}".format(args.background, searchp,bg_ids))
            sys.exit(1)
        else:
            logger.info("Background: {}".format(bg_cls))
            [bg_cls] = bg_ids
    
    for id in resolve(ont1,args.ids,searchp):

        sample = aset1.query([id],[])
        print("Gene set class:{} Gene set: {}".format(id, sample))
        bg = None
        if bg_cls is not None:
            bg = aset1.query([bg_cls],[])
            print("BACKGROUND SUBJECTS: {}".format(bg))

        rs = aset2.enrichment_test(sample, bg, threshold=args.pthreshold, labels=True)
        print("RESULTS: {} < {}".format(len(rs), args.pthreshold))
        for r in rs:
            print(str(r))

def resolve(ont, names, searchp):
    return ont.resolve_names(names,
                              is_remote = searchp.find('x') > -1,
                              is_partial_match = searchp.find('p') > -1,
                              is_regex = searchp.find('r') > -1)
    

    
if __name__ == "__main__":
    main()



import click
import json
import os
import yaml
import requests
import gzip
import urllib
import shutil
import re
import glob
import logging
import copy

import yamldown

from functools import wraps

# from ontobio.util.user_agent import get_user_agent
from ontobio.ontol_factory import OntologyFactory
from ontobio import ontol
from ontobio.io.gafparser import GafParser
from ontobio.io.gpadparser import GpadParser
from ontobio.io.assocwriter import GafWriter
from ontobio.io.assocwriter import GpadWriter
from ontobio.io import assocparser
from ontobio.io import gafgpibridge
from ontobio.io import entitywriter
from ontobio.rdfgen import relations

from typing import Dict, Set

logger = logging.getLogger("INFER")
logger.setLevel(logging.WARNING)

MF = "GO:0003674"
ENABLES = "enables"
HAS_PART = "BFO:0000051"

__ancestors_cache = dict()

@click.group()
@click.option("--log", "-L", type=click.Path(exists=False))
def cli(log):
    global logger
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    logger.addHandler(console_handler)
    if log:
        click.echo("Setting up logging to {}".format(log))
        logfile_handler = logging.FileHandler(log, mode="w")
        logfile_handler.setLevel(logging.INFO)
        logger.addHandler(logfile_handler)
        logger.setLevel(logging.INFO)

@cli.command()
@click.option("--ontology", "-o", "ontology_path", type=click.Path(exists=True), required=True)
@click.option("--target", "-t", type=click.File("w"), required=True)
@click.option("--gaf", "-g", type=click.File("r"), required=True)
def infer(ontology_path, target, gaf):
    ontology_graph = ontology(ontology_path)

    writer = GafWriter(file=target)
    assoc_generator = gafparser_generator(ontology_graph, gaf)
    line_count = 0
    for association in assoc_generator:
        if association["relation"]["id"] != ENABLES:
            continue
            # Skip all non enables annotations

        inferred_associations = materialize_inferences(ontology_graph, association)
        if len(inferred_associations) > 0:
            click.echo("Materialized {} associations".format(len(inferred_associations)))

        for inferred in inferred_associations:
            writer.write_assoc(inferred)

        line_count += 1
        if line_count % 100 == 0:
            click.echo("Processed {} lines".format(line_count))

@cli.command()
@click.option("--ontology", "-o", "ontology_path", type=click.Path(exists=True), required=True)
@click.option("--relation", "-r", required=True)
@click.option("--allowed-trees", multiple=True, default=["biological_process", "molecular_function", "cellular_component"])
def termable(ontology_path, relation, allowed_trees):
    ontology_graph = ontology(ontology_path)

    accum = dict()
    for term in ontology_graph.nodes():
        if term.split(":")[0] != "GO":
            continue

        go_tree = [d["val"] for d in ontology_graph.node(term)["meta"]["basicPropertyValues"] if d["pred"] == "OIO:hasOBONamespace"]
        if len(go_tree) > 0 and go_tree[0] not in allowed_trees:
            continue

        ns = neighbor_by_relation(ontology_graph, term, relation)
        if len(ns) > 0:
            accum[term] = ns

    click.echo(json.dumps(accum, indent=4))

    desc = []
    for term in accum.keys():
        if ontology_graph.children(term, relations=["subClassOf"]):
            desc.append(term)

    click.echo(desc)

def ontology(path) -> ontol.Ontology:
    click.echo("Loading ontology from {}...".format(path))
    return OntologyFactory().create(path, ignore_cache=True)

def ancestors(term: str, ontology: ontol.Ontology, cache) -> Set[str]:
    click.echo("Computing ancestors for {}".format(term))
    if term == MF:
        click.echo("Found 0")
        return set()

    if term not in cache:
        anc = set(ontology.ancestors(term, relations=["subClassOf"], reflexive=True))
        cache[term] = anc
        click.echo("Found {} (from adding to cache: {} terms added)".format(len(anc), len(cache)))
    else:
        anc = cache[term]
        click.echo("Found {} (from cache)".format(len(anc)))

    return anc

def gafparser_generator(ontology_graph: ontol.Ontology, gaf_file):
    config = assocparser.AssocParserConfig(
        ontology=ontology_graph,
    )
    parser = GafParser(config=config)

    return parser.association_generator(gaf_file, skipheader=True)

def neighbor_by_relation(ontology_graph: ontol.Ontology, term, relation):
    return ontology_graph.parents(term, relations=[relation])

def transform_relation(mf_annotation, new_mf, ontology_graph):
    new_annotation = copy.deepcopy(mf_annotation)
    new_annotation["object"]["id"] = new_mf
    return new_annotation

def materialize_inferences(ontology_graph: ontol.Ontology, annotation):
    materialized_annotations = [] #(gp, new_mf)

    mf = annotation["object"]["id"]
    gp = annotation["subject"]["id"]
    global __ancestors_cache
    mf_ancestors = ancestors(mf, ontology_graph, __ancestors_cache)

    # if mf_ancestors:
    #     logger.info("For {term} \"{termdef}\":".format(term=mf, termdef=ontology_graph.label(mf)))
    messages = []

    for mf_anc in mf_ancestors:
        has_part_mfs = neighbor_by_relation(ontology_graph, mf_anc, HAS_PART)

        # if has_part_mfs:
        #     logger.info("\tHas Parent --> {parent} \"{parentdef}\"".format(parent=mf_anc, parentdef=ontology_graph.label(mf_anc)))
        if has_part_mfs:
            messages.append((gp, mf, mf_anc, has_part_mfs))


        for new_mf in has_part_mfs:
            # logger.info("\t\thas_part --> {part} \"{partdef}\"".format(part=new_mf, partdef=ontology_graph.label(new_mf)))

            new_annotation = transform_relation(annotation, new_mf, ontology_graph)
            materialized_annotations.append(new_annotation)


    messages = [ message for message in messages if message[3] ] # Filter out empty has_parts
    for message in messages:
        logger.info("\nFor {gp} -> {term} \"{termdef}\":".format(gp=message[0], term=message[1], termdef=ontology_graph.label(message[1])))
        logger.info("\tHas Parent --> {parent} \"{parentdef}\"".format(parent=message[1], parentdef=ontology_graph.label(message[1])))
        for part in message[3]:
            logger.info("\t\t has_part --> {part} \"{partdef}\"".format(part=part, partdef=ontology_graph.label(part)))
    
    return materialized_annotations



if __name__ == "__main__":
    cli()
